#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <assert.h>
#include <inttypes.h>
#include <sys/types.h>
#include <unistd.h>

#include "tl.h"
#include "x86_sync_defns.h"
#include "logger.h"


#define TLRand(sa) MarsagliaXOR(sa)

#define IS_TX_VALID(x) TxValid(x)
/*
#define IS_TX_VALID(x) ({ \
	int _qwe_ = TxValid(x); \
	ASSERT(_qwe_); \
	_qwe_; \
	})
*/
//-----------------------------
typedef struct _list_t {
    int val;
    struct _list_t *next;
    struct _list_t *prev;
}
list_t;

list_t * tr_list_head=NULL;
int tr_list_size = 0;

int ins_del_thr_launch=0;
int ins_thr_launch=0;
int del_thr_launch=0;
int num_worker_threads;
int num_checker_threads = 1;
int num_rogue_ws_threads = 0;


volatile int can_stop=0;
int seed ;
unsigned int sleep_time;
int pct_ins;
int pct_upd;
int pct_del;
int pct_get;
int init_size;
//-----------------------------

void list_init(Thread*Self,int size) {
    //int seed = ((int) &seed) + RDTICK() ;
    //seed |=1 ;

    if(!IS_TX_VALID(Self))
        return;

    if(size==0) {
        TXSTV(tr_list_head,(intptr_t)NULL);
        if(!IS_TX_VALID(Self))
            return;
        return ;
    }

    // we have at least one elem
    list_t * head = (list_t*)malloc(sizeof(list_t));
    TXSTV(tr_list_head,(intptr_t)head);
    if(!IS_TX_VALID(Self))
        return;
    TXSTF(head,val,0);
    if(!IS_TX_VALID(Self))
        return;
    TXSTF(head,next,NULL);
    if(!IS_TX_VALID(Self))
        return;
    TXSTF(head,prev,NULL);
    if(!IS_TX_VALID(Self))
        return;

    list_t * put = head;
    list_t * prev;

    int i;
    for(i=1;i<size;i++) {
        TXSTF(put,next,(intptr_t)(list_t*)malloc(sizeof(list_t)));
        if(!IS_TX_VALID(Self))
            return;
        prev = put;
        put = (list_t*)TXLDF(put,next);
        if(!IS_TX_VALID(Self))
            return;
        TXSTF(put,prev,prev);
        if(!IS_TX_VALID(Self))
            return;
        TXSTF(put,val,i);
        if(!IS_TX_VALID(Self))
            return;
    }
    TXSTF(put,next,NULL);
    if(!IS_TX_VALID(Self))
        return;

    TXSTV(tr_list_size,size);
    if(!IS_TX_VALID(Self))
        return;

}

int list_check(Thread * Self) {
    list_t * ptr=NULL;
    int num;

    int list_size=TXLDV(tr_list_size);
    if(!IS_TX_VALID(Self))
        return -1;

    list_t * head = (list_t*)TXLDV(tr_list_head);
    if(!IS_TX_VALID(Self))
        return -1;

    assert(head==NULL||(list_t*)TXLDF(head,prev)==NULL);
    if(!IS_TX_VALID(Self))
        return -1;

    num=0;
    ptr=head;
    while(ptr!=NULL) {
        list_t* next = (list_t*)TXLDF(ptr,next);
        if(!IS_TX_VALID(Self))
            return -1;

        if(next!=NULL) {

            list_t * tmp_this = (list_t*)TXLDF(next,prev);
            if(!IS_TX_VALID(Self))
                return -1;
            assert(tmp_this==ptr);

        }
        int val = TXLDF(ptr,val);
        assert(val>=0);


        sched_yield();



        list_t* next2 = (list_t*)TXLDF(ptr,next);
        if(!IS_TX_VALID(Self))
            return -1;

        if(next!=next2) {
            printf("XXXXX num = %d, ptr=%lx, next=%lx, next2=%lx - %lx %lx %lx %lx %lx %lx\n",
                   num,(long)(uintptr_t)ptr,
                   (long)(uintptr_t)next,(long)(uintptr_t)next2,
                   (long)(uintptr_t)ptr->next, (long)(uintptr_t)ptr->prev,
                   (long)(uintptr_t)next->next, (long)(uintptr_t)next->prev,
                   (long)(uintptr_t)next2->next, (long)(uintptr_t)next2->prev);
        }

        ASSERT(next==next2);

        ptr=next;
        num++;
    }

    //printf("num=%d list_size==%d\n",num,list_size);
    assert(num==list_size);
    assert(ptr==NULL);
    return num;

}//list_check

void list_set(Thread * Self) {
    list_t * ptr=NULL;
    int num;
    int val = TLRand(&seed);

    list_t * head = (list_t*)TXLDV(tr_list_head);
    if(!IS_TX_VALID(Self))
        return;

    num=0;
    ptr=head;
    while(ptr!=NULL) {

        TXSTF(ptr,val,val++);
        if(!IS_TX_VALID(Self))
            return;

        ptr = (list_t*)TXLDF(ptr,next);
        if(!IS_TX_VALID(Self))
            return;

        num++;
    }

    assert(ptr==NULL);

}

// only returns NULL when list is empty
list_t * _goto(Thread * Self,int n) {
    int sz = TXLDV(tr_list_size);
    if(!IS_TX_VALID(Self))
        return NULL;

    list_t * head =(list_t*)TXLDV(tr_list_head);
    if(!IS_TX_VALID(Self))
        return NULL;

    if(head==NULL) {
        assert(sz==0);
        return NULL;
    }
    assert(n<sz);

    list_t * ptr = head;
    int num=0;

    while( num<n) {
        //printf("num=%d\n",num);
        ptr=(list_t*)TXLDF(ptr,next);
        if(!IS_TX_VALID(Self))
            return NULL;
        num++;
    }

    return ptr;


}

void _insert(Thread * Self, list_t * new) {
    int pos = TLRand(&seed);
    int sz = TXLDV(tr_list_size);
    if(!IS_TX_VALID(Self))
        return;

    if(sz==0) {
        pos=0;
    } else {
        pos=pos % sz;
    }
    //printf("ins %d - sz = %d\n", pos,sz);

    list_t * ptr;
    ptr = _goto(Self,pos);
    if(!IS_TX_VALID(Self))
        return;

    new->val=pos;

    if(ptr!=NULL) {
        // list is not empty
        new->next=(list_t*)TXLDV(ptr);
        if(!IS_TX_VALID(Self))
            return;
        new->prev=(list_t*)TXLDF(ptr,prev);
        if(!IS_TX_VALID(Self))
            return;
    } else {
        //list is empty
        //        printf("pos=%d, new = %x, next=%x\n", pos,(intptr_t)new,(intptr_t)TXLDF(new,next));
        new->next=NULL;
        //        printf("pos=%d, new = %x, next=%x\n", pos,(intptr_t)new,(intptr_t)TXLDF(new,next));
        //        if(!IS_TX_VALID(Self))
        //            return;
        new->prev=NULL;
        //        if(!IS_TX_VALID(Self))
        //            return;
    }
    if(pos==0) {
        // either pos was set to zero or list is empty
        TXSTV(tr_list_head,(intptr_t)new);
        if(!IS_TX_VALID(Self))
            return;
    }

    if(new->next!=NULL) {
        TXSTF(new->next,prev,new);
        if(!IS_TX_VALID(Self))
            return;
    }

    if(new->prev!=NULL) {
        TXSTF(new->prev,next,new);
        if(!IS_TX_VALID(Self))
            return;

    }

    //    TXSTV(tr_list_size,TXLDV(tr_list_size)+1);
    TXSTV(tr_list_size,sz+1);
    if(!IS_TX_VALID(Self))
        return;
}

void _update(Thread * Self) {
    int pos = TLRand(&seed);
    int sz = TXLDV(tr_list_size);
    if(!IS_TX_VALID(Self))
        return;

    if(sz==0) {
        pos=0;
    } else {
        pos=pos % sz;
    }
    //printf("upd %d\n", pos);
    if(!IS_TX_VALID(Self))
        return;

    list_t * ptr;

    ptr = _goto(Self,pos);
    if(!IS_TX_VALID(Self))
        return;

    if(ptr!=NULL) {
        TXSTF(ptr,val,pos);
        if(!IS_TX_VALID(Self))
            return;
    } else {
        assert(sz==0);
    }
}

void ReleaseNode(Thread * Self, list_t* node) {
    TxSterilize(Self,node,sizeof(list_t)/sizeof(intptr_t));

    node->next=(list_t*)0xFFFF;
    node->prev=(list_t*)0xFFFF;
    node->val=-1;
    free(node);
}

list_t * _delete(Thread * Self) {
    int pos = TLRand(&seed);
    int sz = TXLDV(tr_list_size);
    if(!IS_TX_VALID(Self))
        return NULL;

    if(sz==0) {
        pos=0;
    } else {
        pos=pos % sz;
    }
    //printf("del %d\n", pos);
    if(!IS_TX_VALID(Self))
        return NULL;

    list_t * ptr;

    ptr = _goto(Self,pos);
    if(!IS_TX_VALID(Self))
        return NULL;

    if(ptr!=NULL) {
        list_t * next = (list_t*)TXLDF(ptr,next);
        if(!IS_TX_VALID(Self))
            return NULL;
        list_t * prev = (list_t*)TXLDF(ptr,prev);
        if(!IS_TX_VALID(Self))
            return NULL;

        //list is not empty
        if(next!=NULL) {
            //not the last element
            TXSTF(next,prev,prev);
            if(!IS_TX_VALID(Self))
                return NULL;
        } else {
            // it is the last element
            if(!IS_TX_VALID(Self))
                return NULL;
        }


        if(prev!=NULL) {
            //not the first element
            TXSTF(prev,next,next);
            if(!IS_TX_VALID(Self))
                return NULL;
        } else {
            //it is the first element
            TXSTV(tr_list_head,(intptr_t)next);
            if(!IS_TX_VALID(Self))
                return NULL;
        }

        TXSTV(tr_list_size,sz-1);
        if(!IS_TX_VALID(Self))
            return NULL;

        return ptr;
    } else {
        //list IS empty
        assert(TXLDV(tr_list_size)==0);
        if(!IS_TX_VALID(Self))
            return NULL;

        return NULL;

    }
}

int _get(Thread * Self) {
    int pos = TLRand(&seed);
    int sz = TXLDV(tr_list_size);
    if(!IS_TX_VALID(Self))
        return -1;

    if(sz==0) {
        pos=0;
    } else {
        pos=pos % sz;
    }
    list_t * ptr;

    ptr = _goto(Self,pos);
    if(!IS_TX_VALID(Self))
        return -1;
    if(ptr==NULL)
        return -1;

    int a = TXLDF(ptr,val);
    if(!IS_TX_VALID(Self))
        return -1;

    return a;
}

void * checker_thr(void * args) {
    Thread * Self ;
    Self = TxNewThread() ;

    printf("%d: checker_thr\n", Self->UniqID);

    int checker_successfull_commits=0;
    int checker_failed_commits=0;

    while(!can_stop) {
        int res=0;
        TxStart (Self, 0);
        res=list_check(Self);
        if (TxCommit(Self)) {
            checker_successfull_commits++;
        } else {
            checker_failed_commits++;
        }
        //if((checker_successfull_commits+checker_failed_commits) % 10 ==0){
        //printf("res = %d\n",res);
        //}
        //sched_yield();
    }
    printf("%d: checker_successfull_commits=%d\n", Self->UniqID, checker_successfull_commits);
    printf("%d: checker_failed_commits=%d\n", Self->UniqID, checker_failed_commits);
    TxEndThread(Self);

    return NULL;
}//checker_thr

void * rogue_ws_thr(void * args) {
    Thread * Self ;
    Self = TxNewThread () ;

    printf("%d: rogue_ws_thr\n", Self->UniqID);

    int rogue_ws_successfull_commits=0;
    int rogue_ws_failed_commits=0;

    while(!can_stop) {
        //    	Thread * Self ;
        //        Self = TxNewThread () ;
        //        printf("%d: rogue_ws_thr\n", Self->UniqID);
        TxStart (Self, 0);
        list_set(Self);
        if (TxCommit(Self)) {
            rogue_ws_successfull_commits++;
        } else {
            rogue_ws_failed_commits++;
        }
        //        printf("%d: rogue_ws commited\n", Self->UniqID);
        //        TxEndThread(Self);
        sched_yield();
    }
    printf("%d: rogue_ws_successfull_commits=%d\n", Self->UniqID, rogue_ws_successfull_commits);
    printf("%d: rogue_ws_failed_commits=%d\n", Self->UniqID, rogue_ws_failed_commits);
    TxEndThread(Self);

    return NULL;
}//rogue_ws_thr


void * worker_thr(void * args) {
    Thread * Self ;
    Self = TxNewThread () ;

    printf("%d: worker_thr\n", Self->UniqID);

    int worker_successfull_commits=0;
    int worker_failed_commits=0;


    while(!can_stop) {
        int val = TLRand(&seed)%100 ;
        list_t * deleteable_node =NULL;


        //printf("val=%d\n", val);
        if(val>=0 && val<pct_ins) {
            //            printf("ins\n");
            list_t * new;
            new=(list_t*)malloc(sizeof(list_t));
            //printf("new %x\n", (uintptr_t)new);
            TxStart (Self, 0) ;
            _insert(Self, new);
        } else if(val>=pct_ins && val<pct_ins+pct_upd) {
            //          printf("upd\n");
            TxStart (Self, 0) ;
            _update(Self);
        } else if(val>=pct_ins+pct_upd && val<pct_ins+pct_upd+pct_del) {
            //        printf("del\n");
            TxStart (Self, 0) ;
            deleteable_node = _delete(Self);
        } else if(val>=pct_ins+pct_upd-pct_del && val<pct_ins+pct_upd+pct_del+pct_get) {
            //      printf("get\n");
            TxStart (Self, 0) ;
            _get(Self);
        } else {
            assert(0);
        }

        if (TxCommit(Self)) {
            worker_successfull_commits++;
            if(deleteable_node) {
                //printf("deleted %x\n",(uintptr_t)deleteable_node);
                ReleaseNode(Self,deleteable_node);

            }
        } else {
            worker_failed_commits++;
        }

        //		if((worker_successfull_commits+worker_failed_commits) % 10000 ==0){
        //			printf("alive\n");
        //		}

        //sched_yield();
    }
    printf("%d: worker_successfull_commits=%d\n", Self->UniqID, worker_successfull_commits);
    printf("%d: worker_failed_commits=%d\n", Self->UniqID, worker_failed_commits);
    TxEndThread(Self);


    return NULL;
}//worker_thr



void * del_ins_thr(void * args) {
    Thread * Self ;
    Self = TxNewThread () ;

    while(!can_stop) {
        list_t * deleteable_node =NULL;

        TxStart (Self, 0) ;
        deleteable_node = _delete(Self);
        if (TxCommit(Self)&& deleteable_node) {
            printf("deleted %lx\n",(long)(uintptr_t)deleteable_node);
            ReleaseNode(Self,deleteable_node);
        }

        list_t * new;
        new=(list_t*)malloc(sizeof(list_t));
        printf("new %lx\n", (long)(uintptr_t)new);
        TxStart (Self, 0) ;
        _insert(Self, new);
        TxCommit(Self);

        sched_yield();
    }
    TxEndThread(Self);
    return NULL;
}//del_ins_thr

void * ins_thr(void * args) {
    Thread * Self ;
    Self = TxNewThread () ;

    while(!can_stop) {
        list_t * new;
        new=(list_t*)malloc(sizeof(list_t));
        TxStart (Self, 0) ;
        _insert(Self,new);
        TxCommit(Self);
        //sched_yield();
    }
    TxEndThread(Self);
    return NULL;
}

void * del_thr(void * args) {
    Thread * Self ;
    Self = TxNewThread () ;

    while(!can_stop) {
        list_t * deleteable_node =NULL;
        TxStart (Self, 0) ;
        deleteable_node = _delete(Self);
        if (TxCommit(Self)&& deleteable_node) {
            ReleaseNode(Self,deleteable_node);
        }
        //sched_yield();
    }
    TxEndThread(Self);
    return NULL;
}

//-----------------------------

int main(int argc, char *argv[]) {
    seed= ((int) &seed) + RDTICK() ;
    seed |=1 ;

    if(argc!=8) {
        printf("syntax: bin sleep_time num_worker_threads pct_ins pct_upd pct_del pct_get init_size\n");
        if(argc!=1 && argc!=8)
            exit(1);
    }

    if(argc==8) {
        sleep_time = atoi(argv[1]);
        num_worker_threads = atoi(argv[2]);
        pct_ins = atoi(argv[3]);
        pct_upd = atoi(argv[4]);
        pct_del = atoi(argv[5]);
        pct_get = atoi(argv[6]);
        init_size = atoi(argv[7]);
    } else {
        sleep_time = 2;
        num_worker_threads = 2;
        pct_ins = 25;
        pct_upd = 25;
        pct_del = 25;
        pct_get = 25;
        init_size = 30;
    }
    assert(pct_ins+pct_upd+pct_del+pct_get==100);


    //TxOnce();

    Thread * Self ;
    Self = TxNewThread() ;
    TxStart (Self, 0) ;
    list_init(Self,init_size);
    list_check(Self);
    if (!TxCommit(Self)) {
        assert(0);
    }
    TxStart (Self, 0) ;
    list_check(Self);
    if (!TxCommit(Self)) {
        assert(0);
    }

    TxEndThread(Self);


    pthread_t del_ins_t;
    if(ins_del_thr_launch)
        pthread_create(&del_ins_t,NULL,del_ins_thr,NULL);

    pthread_t ins_t;
    if(ins_thr_launch)
        pthread_create(&ins_t,NULL,ins_thr,NULL);

    pthread_t del_t;
    if(del_thr_launch)
        pthread_create(&del_t,NULL,del_thr,NULL);


    pthread_t *worker_thr_a=(pthread_t*)malloc(sizeof(pthread_t)*num_worker_threads);
    pthread_t *checker_thr_a=(pthread_t*)malloc(sizeof(pthread_t)*num_checker_threads);
    pthread_t *rogue_ws_thr_a=(pthread_t*)malloc(sizeof(pthread_t)*num_rogue_ws_threads);


    int i=0;
    for(i=0; i<num_checker_threads;i++) {
        pthread_create(&(checker_thr_a[i]),NULL,checker_thr,NULL);
    }
    for(i=0; i<num_worker_threads;i++) {
        pthread_create(&(worker_thr_a[i]),NULL,worker_thr,NULL);
    }
    for(i=0; i<num_rogue_ws_threads;i++) {
        pthread_create(&(rogue_ws_thr_a[i]),NULL,rogue_ws_thr,NULL);
    }

    sleep(sleep_time);
    can_stop=1;

    for(i=0; i<num_worker_threads;i++) {
        pthread_join(worker_thr_a[i],NULL);
        printf("worker_join\n");
    }
    printf("end of worker_join's\n");

    for(i=0; i<num_checker_threads;i++) {
        pthread_join(checker_thr_a[i],NULL);
        printf("checker_join\n");
    }
    for(i=0; i<num_rogue_ws_threads;i++) {
        pthread_join(rogue_ws_thr_a[i],NULL);
        printf("rogue_ws_join\n");
    }


    if(ins_del_thr_launch)
        pthread_join(del_ins_t,NULL);
    if(ins_thr_launch)
        pthread_join(ins_t,NULL);
    if(del_thr_launch)
        pthread_join(del_t,NULL);


    //TxShutdown() ;

    free(worker_thr_a);
    free(checker_thr_a);
    free(rogue_ws_thr_a);

    return 0;
}
