#ifndef X86_SYNC_DEFNS_H_
#define X86_SYNC_DEFNS_H_


#define CAS32_internal(_a, _o, _n)                         \
({ __typeof__(_o) __o = _o;                                \
   __asm__ __volatile__(                                   \
       "lock cmpxchg %3,%1"                                \
       : "=a" (__o), "=m" (*(volatile unsigned int *)(_a)) \
       :  "0" (__o), "r" (_n) );                           \
   __o;                                                    \
})

// ONLY WORKS WITH x86_64 
#define CAS64_internal(_a, _o, _n)                               \
({ __typeof__(_o) __o = _o;                                      \
   __asm__ __volatile__(                                         \
       "lock cmpxchg %3,%1"                                      \
       : "=a" (__o), "=m" (*(volatile unsigned long long *)(_a)) \
       :  "0" (__o), "r" (_n) );                                 \
   __o;                                                          \
})

/*
//equiv to atomic{ ret_val=*_a ; *_a=_n ; }
#define FAS32_internal(_a, _n)                             \
({ __typeof__(_n) __o;                                     \
   __asm__ __volatile__(                                   \
       "lock xchg %0,%1"                                   \
       : "=r" (__o), "=m" (*(volatile unsigned int *)(_a)) \
       :  "0" (_n) );                                      \
   __o;                                                    \
})

// ONLY WORKS WITH x86_64 
//equiv to atomic{ ret_val=*_a ; *_a=_n ; }
#define FAS64_internal(_a, _n)                                   \
({ __typeof__(_n) __o;                                           \
   __asm__ __volatile__(                                         \
       "lock xchg %0,%1"                                         \
       : "=r" (__o), "=m" (*(volatile unsigned long long *)(_a)) \
       :  "0" (_n) );                                            \
   __o;                                                          \
})
*/

// ONLY WORKS WITH x86_32 
#define CAS8B_internal(_a, _o, _n)                               \
({ __typeof__(_o) __o = _o;                                      \
   __typeof__(_n) __n = (_n);                                    \
   __typeof__(_n) _nh = (__n >> 32);                             \
   __typeof__(_n) _nl = (__n);                                   \
   __asm__ __volatile__(                                         \
       "movl %3, %%ecx;"                                         \
       "movl %4, %%ebx;"                                         \
       "lock cmpxchg8b %1"                                       \
       : "=A" (__o), "=m" (*(volatile unsigned long long *)(_a)) \
       : "0" (__o), "m" (_nh), "m" (_nl)                         \
       : "ebx", "ecx" );                                         \
   __o;                                                          \
})

// _n = LOAD64(_a) <=> n <= *_a
#define LOAD_64_sse(_a)                             \
({	__typeof__(_a) __a = (_a);                      \
	_u64 __n;                                       \
    __asm__ __volatile__(                           \
       "movl %1, %%eax;"                            \
       "movlps (%%eax), %%xmm0;"                    \
       "movlps %%xmm0, %0"                          \
       : "=m" (__n)                                 \
       : "m" ((volatile unsigned long long *)(__a)) \
       : "%xmm0", "%eax" );                         \
__n;})

// STORE64(_a, _n) <=> *_a <= _n
#define STORE_64_sse(_a, _n)                         \
({	__typeof__(_a) __a = (_a);                       \
	_u64 __n = (_n);                                 \
    __asm__ __volatile__(                            \
       "movlps %1, %%xmm0;"                          \
       "movl %0, %%eax;"                             \
       "movlps %%xmm0, (%%eax)"                      \
       : "=m" ((volatile unsigned long long *)(__a)) \
       : "m" (__n)                                   \
       : "%xmm0", "%eax" );                          \
})


//#define X86_32__32bitversionclock
//#define X86_32__64bitversionclock
//#define X86_64

//X86_32
#if defined X86_32
#define CASIO CAS32_internal
#define CASPO CAS32_internal
#define __LDLOCK(a)    (*(a))
#define __STLOCK(l,n) do{*(l) = n ;}while(0)
#define __CAS_VWLOCK CAS32_internal
#define IS_LOCK64_OK(x) (1)
#define RDTICK RDTICK_X86_32
typedef unsigned int vwLock ;

#define MB()  __asm__ __volatile__ ("lock; addl $0,0(%%esp)" : : : "memory")
#define WMB() __asm__ __volatile__ ("" : : : "memory")
#define RMB() MB()

//X86_32 with SSE instructions
#elif defined X86_32_SSE
#define CASIO CAS32_internal
#define CASPO CAS32_internal
#define __LDLOCK(a)    LOAD_64_sse(a)
#define __STLOCK(a,v) STORE_64_sse(a,v)
#define __CAS_VWLOCK CAS8B_internal
#define IS_LOCK64_OK(x) (((HIGH_32BITS & x) == 0ULL) && IS_LOCKED(x))
#define RDTICK RDTICK_X86_32
typedef unsigned long long vwLock;

#define MB()  __asm__ __volatile__ ("lock; addl $0,0(%%esp)" : : : "memory")
#define WMB() __asm__ __volatile__ ("" : : : "memory")
#define RMB() MB()

//X86_64
#elif defined X86_64
#define CASIO CAS32_internal
#define CASPO CAS64_internal
#define __LDLOCK(a)    (*(a))
#define __STLOCK(l,n) do{*(l) = n ;}while(0)
#define __CAS_VWLOCK CAS64_internal
#define IS_LOCK64_OK(x) (1)
#define RDTICK RDTICK_X86_64
typedef unsigned long long vwLock ;  // (Version,LOCKBIT)

#define MB()    asm volatile("mfence":::"memory")
#define WMB()   asm volatile("" ::: "memory")
#define RMB()   asm volatile("lfence":::"memory")

#else 

#error no flag defined

#endif



#define MEMBARSTLD() MB()
#define MEMBARSTST() WMB()
#define MEMBARLDLD() RMB()

//#define PrefetchW(x) (0)
#define PrefetchW(x) __builtin_prefetch(x, 1, 3);

#define RDTICK_X86_32() \
	({ unsigned long long __t; __asm__ __volatile__ ("rdtsc" : "=A" (__t)); __t; })

#define RDTICK_X86_64() \
	({ unsigned long long __t, __tl, __th; __asm__ __volatile__ ("rdtsc" : "=a" (__tl), "=d" (__th)); __t=(__th<<32)|__tl; __t; })

typedef unsigned char      _u8;
typedef unsigned short     _u16;
typedef unsigned int       _u32;
typedef unsigned long long _u64;


#endif /*X86_SYNC_DEFNS_H_*/
